import tqdm
import imageio
import sys
from typing import Tuple, Optional
from .utils import Utils
import cv2
import os
from multiprocessing import Queue
import math


class Writer:
    def __init__(
            self,
            filename: str,
            frames: Queue,
            resolution: Tuple,
            output: Optional[str] = None,
            low_quality: bool = False,
            fps_reduction: int = 1,
            disable_dither: bool = False,
            shit_optimize=False,
            progress_bar: bool = False,
    ):
        self.filename = filename
        self.frames = frames
        self.resolution = resolution
        self.low_quality = low_quality
        self.output = output
        self.disable_dither = disable_dither
        self.shit_optimize = shit_optimize
        self.progress_bar = progress_bar

        data = Utils.get_video_data(self.filename)
        self.fps = data["fps"]
        self.frame_count = data["frame_count"]
        if 0 < fps_reduction <= self.fps:
            self.fps /= fps_reduction
            self.frame_count = math.ceil(self.frame_count / fps_reduction)

        if not self.output:
            self.output = Utils.get_output_name(self.filename)

    def write_gif(self) -> None:
        quantizer = 0
        if self.low_quality and self.disable_dither:
            quantizer = 2
        with imageio.get_writer(f"{self.output}.gif", mode='I', fps=self.fps, quantizer=quantizer) as writer:
            if self.progress_bar:
                pbar = tqdm.tqdm(total=self.frame_count, desc='Writing frames', position=1)
            for i in range(self.frame_count):
                frame = self.frames.get()
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                writer.append_data(frame)
                if self.progress_bar:
                    pbar.update(1)
        if self.progress_bar:
            sys.stdout.write('\n')
            sys.stdout.flush()
            pbar.close()
        if self.shit_optimize:
            Utils.shit_optimize(self.output)

    def write_video(self) -> None:
        while not self.frames:
            pass

        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        writer = cv2.VideoWriter(f"{self.output}.mp4", fourcc, self.fps, self.resolution)

        if self.progress_bar:
            pbar = tqdm.tqdm(total=self.frame_count, desc='Writing frames', position=1)
        for i in range(self.frame_count):
            frame = self.frames.get()
            writer.write(frame)
            if self.progress_bar:
                pbar.update(1)
        if self.progress_bar:
            sys.stdout.write('\n')
            sys.stdout.flush()
            pbar.close()
        writer.release()



from .reader import Reader
from .writer import Writer
from .text_style import TextStyle
from multiprocessing import Process
from typing import Optional, Tuple
import os
from .utils import Utils
import subprocess

__version__ = "2.6.0"


class FileTypeError(Exception):
    pass


class FileDoesNotExistError(Exception):
    pass


def __get_reader_writer(
        filename: str,
        output: Optional[str] = None,
        width: Optional[int] = None,
        fps_reduction: int = 1,
        low_quality: bool = False,
        disable_dither: bool = False,
        shit_optimize: bool = False,
        text: str = "",
        text_style: TextStyle = TextStyle.TOP,
        progress_bar: bool = False
) -> Tuple[Reader, Writer]:
    reader = Reader(filename, output, width, fps_reduction, disable_dither, shit_optimize, text, text_style, progress_bar)
    writer = Writer(filename, reader.frames, reader.resolution, output, low_quality, fps_reduction, disable_dither, shit_optimize,
                    progress_bar)
    return reader, writer


def video2video(
        filename: str,
        output: Optional[str] = None,
        width: Optional[int] = None,
        fps_reduction: int = 1,
        low_quality: bool = False,
        disable_dither: bool = False,
        shit_optimize: bool = False,
        text: str = "",
        text_style: TextStyle = TextStyle.TOP,
        progress_bar: bool = False
) -> None:
    if not os.path.exists(filename):
        raise FileDoesNotExistError(f"File '{filename}' does not exist.")

    reader, writer = __get_reader_writer(filename, output, width, fps_reduction, low_quality, disable_dither, shit_optimize, text,
                                         text_style, progress_bar)

    try:
        p = Process(target=reader.read_video, args=())
        p.start()
        writer.write_video()
        p.join()
    except KeyboardInterrupt:
        p.terminate()


def gif2gif(
        filename: str,
        output: Optional[str] = None,
        width: Optional[int] = None,
        fps_reduction: int = 1,
        low_quality: bool = False,
        disable_dither: bool = False,
        shit_optimize: bool = False,
        text: str = "",
        text_style: TextStyle = TextStyle.TOP,
        progress_bar: bool = False
) -> None:
    video2gif(filename, output, width, fps_reduction, low_quality, disable_dither, shit_optimize, False, text, text_style,
              progress_bar)


def video2gif(
        filename: str,
        output: Optional[str] = None,
        width: Optional[int] = None,
        fps_reduction: int = 1,
        low_quality: bool = False,
        disable_dither: bool = False,
        shit_optimize: bool = False,
        is_ffmpeg: bool = False,
        text: str = "",
        text_style: TextStyle = TextStyle.TOP,
        progress_bar: bool = False
) -> None:
    if is_ffmpeg:
        if not output:
            output = Utils.get_output_name(filename)

        video2video(filename, "temp", width, fps_reduction, low_quality, True, shit_optimize, text,
                    text_style, progress_bar)

        command = f'ffmpeg -i temp.mp4 -y {output}.gif'
        if not low_quality:
            command = f'ffmpeg -i temp.mp4 -vf "split[s0][s1];[s0]palettegen[p];[s1][p]paletteuse=dither=floyd_steinberg" -y {output}.gif'
        subprocess.call(command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

        if shit_optimize:
            Utils.shit_optimize(output)

        os.remove("temp.mp4")
        return

    if not os.path.exists(filename):
        raise FileDoesNotExistError(f"File '{filename}' does not exist.")

    reader, writer = __get_reader_writer(filename, output, width, fps_reduction, low_quality, disable_dither, shit_optimize, text,
                                         text_style, progress_bar)

    try:
        p = Process(target=reader.read_video, args=())
        p.start()
        writer.write_gif()
        p.join()
    except KeyboardInterrupt:
        p.terminate()


def gif2video(
        filename: str,
        output: Optional[str] = None,
        width: Optional[int] = None,
        fps_reduction: int = 1,
        low_quality: bool = False,
        disable_dither: bool = False,
        shit_optimize: bool = False,
        text: str = "",
        text_style: TextStyle = TextStyle.TOP,
        progress_bar: bool = False
) -> None:
    if not os.path.exists(filename):
        raise FileDoesNotExistError(f"File '{filename}' does not exist.")
    if filename[-4:] != ".gif":
        raise FileTypeError(f"File '{filename}' is not a gif.")

    reader, writer = __get_reader_writer(filename, output, width, fps_reduction, low_quality, disable_dither, shit_optimize, text,
                                         text_style, progress_bar)

    try:
        p = Process(target=reader.read_video, args=())
        p.start()
        writer.write_video()
        p.join()
    except KeyboardInterrupt:
        p.terminate()



from typing import Dict, Tuple, Optional
import cv2
import more_itertools
import numpy as np
from PIL import ImageFont, ImageDraw, Image
from pygifsicle import gifsicle
import os
from .text_style import TextStyle


class Utils(object):
    def get_output_name(filename: str) -> str:
        return "".join(os.path.basename(filename).split('.')[:-1])
    @staticmethod
    def get_video_data(filename: str) -> Dict:
        cap = cv2.VideoCapture(filename)
        fps = int(cap.get(cv2.CAP_PROP_FPS))
        frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        resolution = (int(cap.get(cv2.CAP_PROP_FRAME_WIDTH)),
                      int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT)))
        cap.release()

        return {'fps': fps, 'frame_count': frame_count, 'resolution': resolution}

    @staticmethod
    def create_text_overlay(
            resolution: Tuple,
            text: str,
            width: Optional[int] = None,
            text_style: TextStyle = TextStyle.TOP
    ) -> Optional[Tuple]:

        frame = np.zeros((resolution[1], resolution[0], 3), np.uint8)

        if width:
            frame = Utils.resize_frame(frame, width)

        if text == "":
            return frame, text_style, ((frame.shape[1], frame.shape[0]),)

        if text == "":
            return frame, (frame.shape[1], frame.shape[0])

        frame = Image.new('RGBA', (frame.shape[1], frame.shape[0]), (0, 0, 0, 0))
        size, text = Utils.__calculate_font_size(frame, text)
        font = ImageFont.truetype("impact.ttf", size)
        draw = ImageDraw.Draw(frame)

        x = frame.size[0] // 2 - draw.textsize(text, font=font)[0] // 2
        y = frame.size[1] // 20
        font = ImageFont.truetype("impact.ttf", size)

        if text_style == TextStyle.CAPTION:
            top_margin = draw.textsize(text, font=font)[1] + y * 2
            new_resolution = (frame.width, frame.height + top_margin)
            new_frame = Image.new(frame.mode, new_resolution, (255, 255, 255))
            new_frame.paste(frame, (0, top_margin))
            frame = new_frame
            draw = ImageDraw.Draw(frame)
            draw.text((x, y // 2), text, font=font, fill=(0, 0, 0))
            return frame, text_style, (new_resolution, top_margin)

        if text_style == TextStyle.BOTTOM:
            y = frame.size[1] - y - draw.textsize(text, font=font)[1]

        offset = size * 3 // 80

        draw.text((x - offset - 1, y - offset - 1),
                  text, font=font, fill=(0, 0, 0))
        draw.text((x + offset + 1, y - offset - 1),
                  text, font=font, fill=(0, 0, 0))
        draw.text((x - offset - 1, y + offset + 1),
                  text, font=font, fill=(0, 0, 0))
        draw.text((x + offset + 1, y + offset + 1),
                  text, font=font, fill=(0, 0, 0))

        draw.text((x, y), text, font=font, fill=(255, 255, 255))
        new_resolution = (frame.width, frame.height)
        return frame, text_style, (new_resolution,)

    @staticmethod
    def cv22pil(frame: np.ndarray) -> Image.Image:
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        return Image.fromarray(frame)

    @staticmethod
    def pil2cv2(pil_image: Image.Image) -> np.ndarray:
        return cv2.cvtColor(np.array(pil_image), cv2.COLOR_RGB2BGR)

    @staticmethod
    def morb_frame(
            frame: np.ndarray,
            text_overlay_image: Optional[Image.Image],
            width: Optional[int] = None,
            text: str = "",
            disable_dither: bool = False,
    ) -> np.ndarray:

        if width:
            frame = Utils.resize_frame(frame, width)
        if text != "":
            frame = Utils.caption_video(frame, text_overlay_image)
        if not disable_dither:
            frame = Utils.quantize_dither_frame(frame)
        return frame

    @staticmethod
    def quantize_dither_frame(frame: np.ndarray) -> np.ndarray:
        frame = Utils.cv22pil(frame)
        frame = frame.convert('P', colors=256, palette=Image.Palette.WEB, dither=Image.Dither.FLOYDSTEINBERG)
        frame = frame.convert("RGB")
        frame = Utils.pil2cv2(frame)
        return frame

    @staticmethod
    def shit_optimize(filename: str) -> None:
        gifsicle(sources=f"{filename}.gif", optimize=True, colors=256)

    @staticmethod
    def resize_frame(frame: np.ndarray, width: int) -> np.ndarray:
        if width < 16:
            return frame
        return cv2.resize(frame, (width, frame.shape[0] * width // frame.shape[1]))

    @staticmethod
    def caption_video(frame: np.ndarray, text_overlay_image: Image.Image) -> np.ndarray:
        return Utils.__overlay_text(frame, text_overlay_image)

    @staticmethod
    def __calculate_font_size(frame: Image.Image, text: str = "", depth: int = 0) -> Tuple[int, str]:
        min_size = frame.size[0] // 32
        max_size = frame.size[0] // 16
        max_text_size = frame.size[0] - max_size

        giant_ass_font = ImageFont.truetype("impact.ttf", 10000)
        draw = ImageDraw.Draw(frame)
        size = 10000 * max_text_size // draw.textsize(text, font=giant_ass_font)[0]
        if size < min_size:
            new_text = text.replace("\n", "").split(" ")
            new_text = [list(x)
                        for x in more_itertools.divide(depth + 2, new_text)]
            new_text = " \n".join(" ".join(x) for x in new_text)
            size, text = Utils.__calculate_font_size(
                frame, new_text, depth + 1)
        elif size > max_size:
            size = max_size

        return size, text

    @staticmethod
    def __overlay_text(frame: np.ndarray, text_overlay_image: Image.Image) -> np.ndarray:
        frame = Utils.cv22pil(frame)
        text_style = text_overlay_image[1]
        if text_style == TextStyle.CAPTION:
            top_margin = text_overlay_image[2][1]
            text_overlay_image[0].paste(frame, (0, top_margin))
            frame = text_overlay_image[0]
            return Utils.pil2cv2(frame)
        frame.paste(text_overlay_image[0], (0, 0), text_overlay_image[0])
        return Utils.pil2cv2(frame)



import click
import figconverter
from figconverter.text_style import TextStyle


@click.version_option(figconverter.__version__, "-v", "--version")
@click.command()
@click.argument('filename', type=click.Path(exists=True))
@click.option('-o', '--output', help='Output filename')
@click.option('-g2v', '--gif2video', default=False, is_flag=True, help='Convert a gif to a video')
@click.option('-w', '--width', help='Width of the gif (must be 16 or bigger)', type=int)
@click.option('-fr', '--fps-reduction', default=1,
              help="Divide fps by this number (must be bigger than 0 and can't be bigger than the original fps)",
              type=int)
@click.option('-lq', '--low-quality', is_flag=True, default=False, help='Prioritize speed and size over qualiry',
              show_default=True)
@click.option('-dd', '--disable-dither', is_flag=True, default=False, help='Apply dither to increase quality increase quality (disables low-quality)')
@click.option('-so', '--shit-optimize', default=False, is_flag=True,
              help='Optimize the gif but change it to 256 colors (requires gifsicle)')
@click.option('-f', '--ffmpeg', is_flag=True, default=False, show_default=True,
              help='Use ffmpeg for higher quality conversion (requires ffmpeg) (enables dithering)')
@click.option('-t', '--text', default="", help='Text to add to the gif')
@click.option('-ts', '--text-style', default=TextStyle.TOP.value,
              type=click.Choice([TextStyle.TOP.value, TextStyle.BOTTOM.value, TextStyle.CAPTION.value]),
              help='Style of text to add to the gif')
def main(filename, gif2video, output, width, fps_reduction, low_quality, disable_dither, shit_optimize, ffmpeg, text, text_style):
    if gif2video:
        if filename[-4:] != ".gif":
            raise click.BadParameter(f"File '{filename}' is not a gif.", param_hint='FILENAME')
        figconverter.gif2video(filename, output, width, fps_reduction, low_quality, disable_dither, shit_optimize, text,
                               TextStyle(text_style), True)
        return

    figconverter.video2gif(filename, output, width, fps_reduction, low_quality, disable_dither, shit_optimize, ffmpeg, text,
                           TextStyle(text_style), True)


if __name__ == '__main__':
    main()



import cv2
from .utils import Utils
import tqdm
import sys
from typing import Optional
from multiprocessing import Queue
import math
from .text_style import TextStyle


class Reader:
    def __init__(
        self, 
        filename: str,
        output: Optional[str] = None,
        width: Optional[int] = None,
        fps_reduction: int = 1,
        disable_dither: bool = False,
        shit_optimize: bool = False,
        text: str = "",
        text_style: TextStyle = TextStyle.TOP,
        progress_bar: bool = False
    ):
        self.filename = filename
        self.output = output
        self.width = width
        self.fps_reduction = fps_reduction
        self.disable_dither = disable_dither
        self.shit_optimize = shit_optimize
        self.text = text
        self.text_style = text_style
        self.progress_bar = progress_bar

        data = Utils.get_video_data(self.filename)
        self.frame_count = data['frame_count']
        self.resolution = data['resolution']

        if self.fps_reduction <= 0 or self.fps_reduction > data["fps"]:
            self.fps_reduction = 1
        self.frame_count = math.ceil(self.frame_count / self.fps_reduction)

        self.text_overlay_image = Utils.create_text_overlay(self.resolution, text, width, text_style)
        self.resolution = self.text_overlay_image[2][0]

        self.frames = Queue(self.frame_count)

    def read_video(self) -> None:
        cap = cv2.VideoCapture(self.filename)
        if self.progress_bar:
            pbar = tqdm.tqdm(total=self.frame_count, desc='Reading and processing frames', position=0)
        i = -1
        while cap.isOpened():
            frame = cap.read()[1]
            if frame is not None:
                i += 1
                if i % self.fps_reduction != 0:
                    continue

                frame = Utils.morb_frame(  # process frames
                    frame, 
                    self.text_overlay_image, 
                    self.width, 
                    self.text,
                    self.disable_dither
                )

                self.frames.put(frame)

                if self.progress_bar:
                    pbar.update(1)
            else:
                break
        
        if self.progress_bar:
            pbar.close()
            sys.stdout.write('\x1b[1A')
            sys.stdout.flush()
        cap.release()

        return



from setuptools import setup
from pathlib import Path
import figconverter

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='fig-converter',
    description='A python package and CLI to turn videos into GIFs and vice-versa.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=figconverter.__version__,
    license='MIT',
    author="blahberi, kamoodi",
    author_email='droplet.org@gmail.com',
    packages=["figconverter"],
    py_modules=["figconverter"],
    url='https://github.com/dropletOrg/FIG',
    install_requires=['opencv-python', 'numpy', 'pillow', 'click', 'tqdm', 'pygifsicle', 'more-itertools', 'imageio'],
    scripts=['fig.py'],
    keywords='convert converter video gif fig pyfig figpy python-fig fig-python py-fig fig-py caption text edit editor cli package command command-line gif-convert gif-converter tool gif-tool gif-editor gif-editor-tool',
)
